#include "SnakeDataManager.h"

#include <QGeoCoordinate>
#include <QMutexLocker>

#include "QGCApplication.h"
#include "SettingsManager.h"
#include "QGCToolbox.h"
#include "WimaSettings.h"
#include "SettingsFact.h"

#include <memory>
#include <shared_mutex>
#include <mutex>

#include "snake.h"
#include "Wima/Snake/SnakeTiles.h"
#include "Wima/Snake/SnakeTilesLocal.h"

#include "comm/ros_bridge/include/ros_bridge.h"

using QVariantList = QList<QVariant>;
using ROSBridgePtr = std::unique_ptr<ros_bridge::ROSBridge>;

using UniqueLock = std::unique_lock<shared_timed_mutex>;
using SharedLock = std::shared_lock<shared_timed_mutex>;

class SnakeImpl : public QObject{
    Q_OBJECT
public:
    SnakeImpl() :
        lineDistance(1*si::meter)
      , minTransectLength(1*si::meter)
      , calcInProgress(false)
      , topicServiceSetupDone(false)
    {
        // ROS Bridge.
        WimaSettings* wimaSettings = qgcApp()->toolbox()->settingsManager()->wimaSettings();
        auto connectionStringFact = wimaSettings->rosbridgeConnectionString();
        auto setConnectionString = [connectionStringFact, this]{
            auto connectionString = connectionStringFact->rawValue().toString();
            if ( ros_bridge::isValidConnectionString(connectionString.toLocal8Bit().data()) ){
                this->pRosBridge.reset(new ros_bridge::ROSBridge(connectionString.toLocal8Bit().data()));
            } else {
                qgcApp()->showMessage("ROS Bridge connection string invalid: " + connectionString);
                this->pRosBridge.reset(new ros_bridge::ROSBridge("localhost:9090"));
            }
        };
        setConnectionString();
        connect(connectionStringFact, &SettingsFact::rawValueChanged, setConnectionString);
    }

    // Private data.
    ROSBridgePtr                pRosBridge;
    bool                        topicServiceSetupDone;
    mutable std::shared_timed_mutex   mutex;

    // Input
    snake::Scenario scenario;
    Length lineDistance;
    Length minTransectLength;
    QList<QGeoCoordinate> mArea;
    QList<QGeoCoordinate> sArea;
    QList<QGeoCoordinate> corridor;


    // Output
    std::atomic_bool            calcInProgress;
    QNemoProgress               qProgress;
    std::vector<int>            progress;
    QNemoHeartbeat              heartbeat;

    QVector<QGeoCoordinate>     waypoints;
    QVector<QGeoCoordinate>     arrivalPath;
    QVector<QGeoCoordinate>     returnPath;
    QGeoCoordinate              ENUorigin;
    QVector<QPointF>            waypointsENU;
    QVector<QPointF>            arrivalPathENU;
    QVector<QPointF>            returnPathENU;
    SnakeTiles                  tiles;
    QVariantList                tileCenterPoints;
    SnakeTilesLocal             tilesENU;
    QVector<QPointF>            tileCenterPointsENU;
    QString                     errorMessage;
};

template<class AtomicType>
class ToggleRAII{
public:
    ToggleRAII(AtomicType &t)
        : _t(t)
    {}

    ~ToggleRAII()
    {
        if ( _t.load() ){
            _t.store(false);
        } else {
            _t.store(true);
        }
    }
private:
    AtomicType &_t;
};



SnakeDataManager::SnakeDataManager(QObject *parent)
    : QThread(parent)
    , pImpl(std::make_unique<SnakeImpl>())

{
}

SnakeDataManager::~SnakeDataManager()
{

}

void SnakeDataManager::setMeasurementArea(const QList<QGeoCoordinate> &measurementArea)
{
    UniqueLock lk(this->pImpl->mutex);
    this->pImpl->mArea = measurementArea;
}

void SnakeDataManager::setServiceArea(const QList<QGeoCoordinate> &serviceArea)
{
    UniqueLock lk(this->pImpl->mutex);
    this->pImpl->sArea = serviceArea;
}

void SnakeDataManager::setCorridor(const QList<QGeoCoordinate> &corridor)
{
    UniqueLock lk(this->pImpl->mutex);
    this->pImpl->corridor = corridor;
}

QNemoProgress SnakeDataManager::progress()
{
    SharedLock lk(this->pImpl->mutex);
    return this->pImpl->qProgress;
}

QNemoHeartbeat SnakeDataManager::heartbeat()
{
    SharedLock lk(this->pImpl->mutex);
    return this->pImpl->heartbeat;
}

bool SnakeDataManager::calcInProgress()
{
    return this->pImpl->calcInProgress.load();
}

Length SnakeDataManager::lineDistance() const
{
    SharedLock lk(this->pImpl->mutex);
    return this->pImpl->lineDistance;
}

void SnakeDataManager::setLineDistance(Length lineDistance)
{
    UniqueLock lk(this->pImpl->mutex);
    this->pImpl->lineDistance = lineDistance;
}

Area SnakeDataManager::minTileArea() const
{
    SharedLock lk(this->pImpl->mutex);
    return this->pImpl->scenario.minTileArea();
}

void SnakeDataManager::setMinTileArea(Area minTileArea)
{
    UniqueLock lk(this->pImpl->mutex);
    this->pImpl->scenario.setMinTileArea(minTileArea);
}

Length SnakeDataManager::tileHeight() const
{
    SharedLock lk(this->pImpl->mutex);
    return this->pImpl->scenario.tileHeight();
}

void SnakeDataManager::setTileHeight(Length tileHeight)
{
    UniqueLock lk(this->pImpl->mutex);
    this->pImpl->scenario.setTileHeight(tileHeight);
}

Length SnakeDataManager::tileWidth() const
{
    SharedLock lk(this->pImpl->mutex);
    return this->pImpl->scenario.tileWidth();
}

void SnakeDataManager::setTileWidth(Length tileWidth)
{
    UniqueLock lk(this->pImpl->mutex);
    this->pImpl->scenario.setTileWidth(tileWidth);
}

bool SnakeDataManager::precondition() const
{
    return true;
}

//!
//! \brief Resets waypoint data.
//! \pre this->_pImpl->mutex must be locked.
void SnakeDataManager::resetWaypointData()
{
    this->pImpl->waypoints.clear();
    this->pImpl->arrivalPath.clear();
    this->pImpl->returnPath.clear();
    this->pImpl->ENUorigin = QGeoCoordinate(0.0,0.0,0.0);
    this->pImpl->waypointsENU.clear();
    this->pImpl->arrivalPathENU.clear();
    this->pImpl->returnPathENU.clear();
    this->pImpl->tiles.polygons().clear();
    this->pImpl->tileCenterPoints.clear();
    this->pImpl->tilesENU.polygons().clear();
    this->pImpl->tileCenterPointsENU.clear();
    this->pImpl->errorMessage.clear();
}

void SnakeDataManager::run()
{
    this->pImpl->calcInProgress.store(true);
    ToggleRAII<std::atomic_bool> tr(this->pImpl->calcInProgress);

    UniqueLock lk(this->pImpl->mutex);
    resetWaypointData();

    if ( !precondition() )
        return;

    if ( this->pImpl->mArea.size() < 3) {
        pImpl->errorMessage = "Measurement area invalid: size < 3.";
        return;
    }
    if ( this->pImpl->sArea.size() < 3) {
        pImpl->errorMessage = "Service area invalid: size < 3.";
        return;
    }


    this->pImpl->ENUorigin = this->pImpl->mArea.front();
    auto &origin = this->pImpl->ENUorigin;
    // Update measurement area.
    auto &mAreaEnu = this->pImpl->scenario.measurementArea();
    auto &mArea = this->pImpl->mArea;
    mAreaEnu.clear();
    for (auto geoVertex : mArea){
        snake::BoostPoint p;
        snake::toENU(origin, geoVertex, p);
        mAreaEnu.outer().push_back(p);
    }

    // Update service area.
    auto &sAreaEnu = this->pImpl->scenario.serviceArea();
    auto &sArea = this->pImpl->sArea;
    sAreaEnu.clear();
    for (auto geoVertex : sArea){
        snake::BoostPoint p;
        snake::toENU(origin, geoVertex, p);
        sAreaEnu.outer().push_back(p);
    }

    // Update corridor.
    auto &corridorEnu = this->pImpl->scenario.corridor();
    auto &corridor = this->pImpl->corridor;
    corridor.clear();
    for (auto geoVertex : corridor){
        snake::BoostPoint p;
        snake::toENU(origin, geoVertex, p);
        corridorEnu.outer().push_back(p);
    }

    if ( !this->pImpl->scenario.update() ){
        pImpl->errorMessage = this->pImpl->scenario.errorString.c_str();
        return;
    }

    // Asynchronously update flightplan.
    qWarning() << "route calculation missing";
    std::string errorString;
    auto future = std::async([this, &errorString]{
        auto alpha = -this->pImpl->scenario.mAreaBoundingBox().angle*degree::degree;
        snake::flight_plan::Transects transects;
        transects.push_back(bg::model::linestring{this->pImpl->scenario.homePositon()});
        bool value = snake::flight_plan::transectsFromScenario(this->pImpl->lineDistance,
                                                             this->pImpl->minTransectLength,
                                                             alpha,
                                                             this->pImpl->scenario,
                                                             this->pImpl->progress,
                                                             transects,
                                                             errorString);
        if ( !value )
            return value;
        snake::flight_plan::Transects transectsRouted;
        snake::flight_plan::Route route;
        value = snake::flight_plan::route(this->pImpl->scenario.joinedArea(),
                                          transects,
                                          transectsRouted,
                                          route,
                                          errorString);
        if ( !value )
            return value;

        // Store waypoints.
        for (auto boostVertex : _pFlightplan->waypoints()){
            QPointF enuVertex{boostVertex.get<0>(), boostVertex.get<1>()};
            QGeoCoordinate geoVertex;
            snake::fromENU(origin, boostVertex, geoVertex);
            pImpl->waypointsENU.push_back(enuVertex);
            pImpl->waypoints.push_back(geoVertex);
        }
        // Store arrival path.
        for (auto boostVertex : _pFlightplan->arrivalPath()){
            QPointF enuVertex{boostVertex.get<0>(), boostVertex.get<1>()};
            QGeoCoordinate geoVertex;
            snake::fromENU(origin, boostVertex, geoVertex);
            pImpl->arrivalPathENU.push_back(enuVertex);
            pImpl->arrivalPath.push_back(geoVertex);
        }
        // Store return path.
        for (auto boostVertex : _pFlightplan->returnPath()){
            QPointF enuVertex{boostVertex.get<0>(), boostVertex.get<1>()};
            QGeoCoordinate geoVertex;
            snake::fromENU(origin, boostVertex, geoVertex);
            pImpl->returnPathENU.push_back(enuVertex);
            pImpl->returnPath.push_back(geoVertex);
        }

        return true;

    });

    // Continue with storing scenario data in the mean time.
    {
        // Get tiles.
        const auto &tiles = this->pImpl->scenario.tiles();
        const auto &centerPoints = this->pImpl->scenario.tileCenterPoints();
        for ( unsigned int i=0; i < tiles.size(); ++i ) {
            const auto &tile = tiles[i];
            SnakeTile geoTile;
            SnakeTileLocal enuTile;
            for ( size_t i = tile.outer().size(); i < tile.outer().size()-1; ++i) {
                auto &p = tile.outer()[i];
                QPointF enuVertex(p.get<0>(), p.get<1>());
                QGeoCoordinate geoVertex;
                snake::fromENU(origin, p, geoVertex);
                enuTile.polygon().points().push_back(enuVertex);
                geoTile.push_back(geoVertex);
            }
            const auto &boostPoint = centerPoints[i];
            QPointF enuVertex(boostPoint.get<0>(), boostPoint.get<1>());
            QGeoCoordinate geoVertex;
            snake::fromENU(origin, boostPoint, geoVertex);
            geoTile.setCenter(geoVertex);
            pImpl->tiles.polygons().push_back(geoTile);
            pImpl->tileCenterPoints.push_back(QVariant::fromValue(geoVertex));
            pImpl->tilesENU.polygons().push_back(enuTile);
            pImpl->tileCenterPointsENU.push_back(enuVertex);
         }
    }

    future.wait();
    // Trying to generate flight plan.
    if ( !future.get() ){
        // error
        for (auto c : _pFlightplan->errorString){
            pImpl->errorMessage.push_back(QChar(c));
        }
    } else {
        // Success!!!

    }
}


